@everywhere function Jacobi_serial(A,N,T)
  B = A
  for t=1:T
      for i=2:(N-1), j=2:(N-1)
          B[i,j] = 0.25 * (A[i-1,j]+A[i+1,j]+A[i,j-1]+A[i,j+1])
      end
      for i=2:(N-1), j=2:(N-1)
          A[i,j] = B[i,j]
      end 
  end
end

N = 10
A = zeros(N,N)
for j=1:N A[1,j]=1 ; A[N, j]=1 end
for i=1:N A[i,1]=1 ; A[i, N]=1 end
T = 1000
Jacobi_serial(A,N,T)

## This function performs one-time-step Jacobi simulation
## in A[x:x+k-1, y:y+k-1]
@everywhere function Jacobi_serial_one_block_one_time_step(A,x,y,k)
  ## We copy a k-by-k block and the elements around it from A to B
  B = A[(x-1):(x-1+k+1),(y-1):(y-1+k+1)]
  ## We do one time of Jacobi, updating the interior of B
  for i=2:(k+1), j=2:(k+1)
          B[i,j] = 0.25 * (A[x-2+i-1,y-2+j]+A[x-2+i+1,y-2+j]+
                           A[x-2+i,y-2+j-1]+A[x-2+i,y-2+j+1])
  end
  ## We copy back the interior of B into A
  for i=2:(k+1), j=2:(k+1)
          A[x-2+i,y-2+j] = B[i,j]
  end 
end

N = 10
A = ones(N,N)
Jacobi_serial_one_block_one_time_step(A,2,2,4)
Jacobi_serial_one_block_one_time_step(A,2,6,4)
Jacobi_serial_one_block_one_time_step(A,6,2,4)
Jacobi_serial_one_block_one_time_step(A,6,6,4)

## The input grid A has N columns and N rows
## We assume that the first and last rows and columns
## have been intialized and should not be updated
@everywhere function Jacobi_parallel(A,N,T)
  for t=1:T
     Jacobi_blockwise_parallel_one_time_step(A,2,2,N-2,512)
  end
end


## Assumptions K writes a power of 2 times a number less or equal then threshold
@everywhere function Jacobi_blockwise_parallel_one_time_step(A,x,y,k,threshold)
     if k <= threshold
        Jacobi_serial_one_block_one_time_step(A,x,y,k)
     else
        k1 = ifloor(k/2)
        k2 = k - k1
        if (k1 != k2) error ("This case is not supported yet") end
        ## CODE IS MISSING HERE
    end
end


N = 10
A = rand(N,N)
SA = convert(SharedArray, A)
T = 100
Jacobi_serial(A,N,T)
Jacobi_parallel(SA,N,T)
A - SA


##N = 2048
##A = zeros(N,N)
##for j=1:N A[1,j]=1 ; A[N, j]=1 end
##for i=1:N A[i,1]=1 ; A[i, N]=1 end
##T = 500
##SA = convert(SharedArray, A)
##Jacobi_parallel(SA,N,T)


N = 2050
A = rand(N,N)
SA = convert(SharedArray, A)
T = 10
@time Jacobi_serial(A,N,T)
@time Jacobi_parallel(SA,N,T)

N = 4098
A = rand(N,N)
SA = convert(SharedArray, A)
T = 10
@time Jacobi_serial(A,N,T)
@time Jacobi_parallel(SA,N,T)

N = 8194
A = rand(N,N)
SA = convert(SharedArray, A)
T = 10
@time Jacobi_serial(A,N,T)
@time Jacobi_parallel(SA,N,T)

N = 16386
A = rand(N,N)
SA = convert(SharedArray, A)
T = 10
@time Jacobi_serial(A,N,T)
@time Jacobi_parallel(SA,N,T)